/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __math_h__
#define __math_h__

#include "fe/core.h"
#include <limits>


#ifdef MODULE_math
#define FE_MATH_PORT FE_DL_EXPORT
#else
#define FE_MATH_PORT FE_DL_IMPORT
#endif

#define	FE_EXPLOIT_MEMCPY			TRUE
#define	FE_BOUNDSCHECK				(FE_CODEGEN<=FE_DEBUG)

#define FE_INVALID_SCALAR(scalar)	(scalar != scalar)

namespace fe
{
const F32 pi=3.1415926f;
const F32 degToRad=fe::pi/180.0f;
#if FE_DOUBLE_REAL
const Real tol=1e-8;
#else
const Real tol=1e-5f;
#endif

extern Real epsilon;
class EpsilonCalc
{
	public:
		EpsilonCalc(void)
		{
			epsilon = 1.0;
			do { epsilon /= 2.0f; }
			while ((Real)(1.0 + (epsilon/2.0)) != 1.0);
		}
};
extern const EpsilonCalc epscalc;

typedef enum
{
	e_xAxis=0x00,
	e_yAxis=0x01,
	e_zAxis=0x02
} Axis;

template<class T>	T minimum(T a,T b)	{ return a>b? b: a; }
template<class T>	T maximum(T a,T b)	{ return a>b? a: b; }

template<class T>	class Matrix3x4;
}

#include "math/Vector.h"
#include "math/Vector2.h"
#include "math/Vector3.h"
#include "math/Vector4.h"
#include "math/Vector_gnu.h"
#include "math/Color.h"
#include "math/Matrix.h"
#include "math/Quaternion.h"
#include "math/Matrix3x4.h"
#include "math/OrthonormalBasis.h"
#include "math/Euler.h"
#include "math/Box.h"
#include "math/DAGNode.h"
#include "math/Barycenter.h"

namespace fe
{

typedef Vector<3,Real> SpatialVector;
typedef Matrix<3,3,Real> SpatialMatrix;

template<typename T,typename U>
inline bool isZero(T a_value,U a_tolerance)
{
	if(fabs(a_value) > a_tolerance)
	{
		return false;
	}
	return true;
}

inline bool isZero(const SpatialVector &a_vec,Real a_tolerance)
{
	if(!isZero(a_vec[0],a_tolerance) || !isZero(a_vec[1],a_tolerance) ||
			!isZero(a_vec[2],a_tolerance))
	{
		return false;
	}
	return true;
}

template<typename T>
inline bool isZero(T a_value)
{
	return isZero(a_value,fe::tol);
}

inline bool isZero(const SpatialVector &a_vec)
{
	return isZero(a_vec,fe::tol);
}

const SpatialVector zeroVector(Real(0),Real(0),Real(0));

}

#include "math/Mass.h"
#include "math/assertMath.h"
#include "math/algorithms.h"

#endif // __math_h__

